import warnings
from io import BytesIO
from zipfile import ZipFile, ZIP_DEFLATED

from openpyxl.xml.constants import COMMENTS_NS
from openpyxl.cell import MergedCell
from openpyxl.comments.comment_sheet import CommentSheet
from openpyxl.packaging.relationship import (
    RelationshipList,
    get_dependents,
    get_rels_path
)

from openpyxl.pivot.table import TableDefinition
from openpyxl.chartsheet import Chartsheet
from openpyxl.worksheet.table import Table
from openpyxl.drawing.spreadsheet_drawing import SpreadsheetDrawing
from openpyxl.xml.constants import (
    ARC_SHARED_STRINGS,
    ARC_CORE,
    ARC_CONTENT_TYPES,
    ARC_WORKBOOK,
    ARC_THEME,
    COMMENTS_NS,
    SHARED_STRINGS,
    EXTERNAL_LINK,
    XLTM,
    XLTX,
    XLSM,
    XLSX,
)
from openpyxl.xml.functions import fromstring
from openpyxl.reader.drawings import find_images

from openpyxl.reader.excel import ExcelReader as _ExcelReader, _find_workbook_part
from .WorksheetReader import WorksheetReader
from .WorkbookParser import WorkbookParser
'''
data_only default to False
'''
class ExcelReader(_ExcelReader):
    def read_workbook(self):
        wb_part = _find_workbook_part(self.package)
        self.parser = WorkbookParser(self.archive, wb_part.PartName[1:], keep_links=self.keep_links)
        self.parser.parse()
        wb = self.parser.wb
        wb._sheets = []
        wb._data_only = self.data_only
        wb._read_only = self.read_only
        wb.template = wb_part.ContentType in (XLTX, XLTM)

        # If are going to preserve the vba then attach a copy of the archive to the
        # workbook so that is available for the save.
        if self.keep_vba:
            wb.vba_archive = ZipFile(BytesIO(), 'a', ZIP_DEFLATED)
            for name in self.valid_files:
                wb.vba_archive.writestr(name, self.archive.read(name))

        if self.read_only:
            wb._archive = self.archive

        self.wb = wb

    def read_worksheets(self):
        comment_warning = """Cell '{0}':{1} is part of a merged range but has a comment which will be removed because merged cells cannot contain any data."""
        for sheet, rel in self.parser.find_sheets():
            if rel.target not in self.valid_files:
                continue

            if "chartsheet" in rel.Type:
                self.read_chartsheet(sheet, rel)
                continue

            rels_path = get_rels_path(rel.target)
            rels = RelationshipList()
            if rels_path in self.valid_files:
                rels = get_dependents(self.archive, rels_path)

            # Used Worksheet and WorksheetReader for all scenarios, 
            # not distinguishing between read_only or not
            fh = self.archive.open(rel.target)
            ws = self.wb.create_sheet(sheet.name)
            ws._rels = rels
            ws_parser = WorksheetReader(ws, fh, self.shared_strings, self.data_only, self.read_only)
            ws_parser.bind_all()

            # assign any comments to cells
            for r in rels.find(COMMENTS_NS):
                src = self.archive.read(r.target)
                comment_sheet = CommentSheet.from_tree(fromstring(src))
                for ref, comment in comment_sheet.comments:
                    try:
                        ws[ref].comment = comment
                    except AttributeError:
                        c = ws[ref]
                        if isinstance(c, MergedCell):
                            warnings.warn(comment_warning.format(ws.title, c.coordinate))
                            continue

            # preserve link to VML file if VBA
            if self.wb.vba_archive and ws.legacy_drawing:
                ws.legacy_drawing = rels[ws.legacy_drawing].target

            for t in ws_parser.tables:
                src = self.archive.read(t)
                xml = fromstring(src)
                table = Table.from_tree(xml)
                ws.add_table(table)

            drawings = rels.find(SpreadsheetDrawing._rel_type)
            for rel in drawings:
                charts, images = find_images(self.archive, rel.target)
                for c in charts:
                    ws.add_chart(c, c.anchor)
                for im in images:
                    ws.add_image(im, im.anchor)

            pivot_rel = rels.find(TableDefinition.rel_type)
            for r in pivot_rel:
                pivot_path = r.Target
                src = self.archive.read(pivot_path)
                tree = fromstring(src)
                pivot = TableDefinition.from_tree(tree)
                pivot.cache = self.parser.pivot_caches[pivot.cacheId]
                ws.add_pivot(pivot)

            ws.sheet_state = sheet.state


import logging
from openpyxl.cell import Cell as _Cell

logger = logging.getLogger(__name__)

class Cell(_Cell):
    __slots__ = ('_cache', '_cache_type')
    def __init__(self, worksheet, row=None, column=None, value=None, style_array=None, cache=None, cache_type=None):
        super().__init__(worksheet, row=row, column=column, value=value, style_array=style_array)
        self._cache = cache
        self._cache_type = cache_type

    @property
    def is_formula(self):
        return self.data_type == 'f'
    
    @property
    def cache(self):
        return self._cache
    
    @property
    def cache_type(self):
        return self._cache_type

    @property
    def data(self):
        # out = self._value
        if self.is_formula:
            if self.cache is not None:
                out = self.cache
            else:
                out = self.parent.parent._compute(self.parent.title, self.coordinate)
            if out == None:
                out = self._value
        else:
            out = self._value
        if self.parent.header is None or self.row <= self.parent.header_row:
            return out
        if self.parent.use_default:
            if out is None:
                try:
                    out = self.parent.header.get_default(self.column)
                except Exception as e:
                    logger.debug(str(e))
            else:
                default_type = self.parent.header.get_type(self.column)
                if default_type is not None:
                    try:
                        out = default_type(out)
                    except Exception as e:
                        if default_type == int:
                            out = 0
                        elif default_type == float:
                            out = 0.0
                        logger.warning(str(e))
        return out
    
    @data.setter
    def data(self, value):
        self._bind_value(value)
    
    @property
    def horizontal(self):
        return tuple(self.parent.cell(self.row, x) for x in range(self.column, self.parent.max_column + 1))
    
    @horizontal.setter
    def horizontal(self, it):
        for idx, item in enumerate(it):
            self.parent.cell(self.row, self.column + idx, item)
    
    @property
    def vertical(self):
        return tuple(self.parent.cell(x, self.column) for x in range(self.row, self.parent.max_row + 1))
    
    @vertical.setter
    def vertical(self, it):
        for idx, item in enumerate(it):
            self.parent.cell(self.row + idx, self.column, item)
    
    @property
    def top(self):
        if self.row == 1:
            raise IndexError(f"cell({self.row}, {self.column}) has no cell above.")
        return self.parent.cell(self.row - 1, self.column)
    
    @property
    def bottom(self):
        return self.parent.cell(self.row + 1, self.column)
    
    @property
    def left(self):
        if self.column == 1:
            raise IndexError(f"cell({self.row}, {self.column}) has not cell on its left.")
        return self.parent.cell(self.row, self.column - 1)
    
    @property
    def right(self):
        return self.parent.cell(self.row, self.column + 1)
    

        
            
            


            



import logging
from openpyxl.cell.read_only import ReadOnlyCell as _ReadOnlyCell, EmptyCell as _EmptyCell
from .Cell import Cell

logger = logging.getLogger(__name__)

class ReadOnlyCell(_ReadOnlyCell):
    __slots__ = ('_cache', '_cache_type')
    def __init__(self, sheet, row, column, value, data_type='n', style_id=0, cache=None, cache_type=None):
        super().__init__(sheet, row, column, value, data_type=data_type, style_id=style_id)
        self._cache = cache
        self._cache_type = cache_type

    @property
    def is_formula(self):
        return Cell.is_formula.__get__(self)
    
    @property
    def cache(self):
        return self._cache
    
    @property
    def cache_type(self):
        return self._cache_type

    @property
    def data(self):
        return Cell.data.__get__(self)
    
    @data.setter
    def data(self, value):
        if self._value is not None:
            raise AttributeError("Cell is read only")
        self._value = value

    @property
    def horizontal(self):
        return Cell.horizontal.__get__(self)
    
    @property
    def vertical(self):
        return Cell.vertical.__get__(self)
    
    @property
    def top(self):
        return Cell.top.__get__(self)
    
    @property
    def bottom(self):
        return Cell.bottom.__get__(self)
    
    @property
    def left(self):
        return Cell.left.__get__(self)
    
    @property
    def right(self):
        return Cell.right.__get__(self)


from typing import Dict, List, Optional, Union, Tuple
import logging
from enum import Enum
# from pyxlsx.Worksheet import Worksheet
from .Cell import Cell

logger = logging.getLogger(__name__)

class InvalidOperationError(Exception):
    pass

class SeriesType(Enum):
    Row = 0
    Column = 1

class Series:
    '''
    Base class for Row and Column
    '''
    def __init__(self, parent, cells, series_type):
        if len(cells) == 0:
            raise ValueError(f"Cannot init Series with empty cells.")
        self._series_type: SeriesType = series_type
        self.parent = parent
        self.cells: Tuple[Cell] = cells
        self.cell_dict: Dict[int, Cell] = self._gen_cell_dict()
    
    def _gen_cell_dict(self):
        if self.series_type == SeriesType.Row:
            return { c.column: c for c in self.cells }
        else:
            return { c.row: c for c in self.cells }
    
    def __len__(self):
        return len(self.cells)
    
    def __iter__(self):
        return iter(self.cells)
    
    def __contains__(self, value):
        for c in self.cells:
            if value == c.data:
                return True
        return False
    
    @property
    def values(self):
        return tuple(c.data for c in self.cells)
    
    @property
    def series_type(self):
        return self._series_type
    
    def __len__(self):
        if self._series_type == SeriesType.Row:
            return self.parent.max_column + 1 - self.parent.min_column
        elif self._series_type == SeriesType.Column:
            return self.parent.max_row + 1 - self.parent.min_row

class ContentRow(Series):
    def __init__(self, parent, cells):
        super().__init__(parent, cells, SeriesType.Row)
        self._min_column: int = self.cells[0].column
        self._max_column: int = self.cells[-1].column
    
    @property
    def row(self):
        return self.cells[0].row

    @property
    def max_column(self):
        return self._max_column
    
    @property
    def min_column(self):
        return self._min_column
    
    def __getitem__(self, key):
        column = self.key_to_column(key)
        try:
            return self.cell_dict[column].data
        except KeyError:
            self.refresh(self.parent.header.max_column - self.max_column)
            return self.cell_dict[column].data
    
    def __setitem__(self, key, value):
        column = self.key_to_column(key)
        self.cell_dict[column].data = value
    
    def refresh(self, increment=0, new_row=None):
        '''
        Refresh cells.
        Args:
            increment: additional columns to include.
            new_row: get cells from different row.
        '''
        if new_row is not None and not isinstance(new_row, int):
            raise TypeError(f"new_row should be of type int or None (got {type(new_row)}.")
        if new_row is not None and new_row <= 0:
            raise ValueError(f"new_row should be positive (got {new_row}).")
        row = new_row or self.row
        self._max_column += increment
        self.cells = tuple(self.parent.cell(row, column) for column in range(self._min_column, self._max_column + 1))
        self.cell_dict = self._gen_cell_dict()

    def append(self, value):
        if value is None:
            return
        self.parent.cell(self.row, self._max_column + 1, value)
        self.refresh(increment=1)
        if not isinstance(self, Header) and self.max_column > self.parent.header.max_column:
            self.parent.header.refresh(self.max_column - self.parent.header.max_column)
    
    def update(self, data_dict, update_header=True):
        for k, v in data_dict.items():
            column = self.parent.header.key_to_column(k)
            if column is not None:
                if column in self.cell_dict:
                    self.cell_dict[column].data = v
                else:
                    self.parent.cell(self.row, column, v)
                    self.refresh(self.parent.header.max_column - self.max_column)
            elif update_header:
                self.parent.header.append(k)
                self.refresh(self.parent.header.max_column - self.max_column)
                self[k] = v
    
    def extend(self, iterable):
        if len(iterable) == 0:
            return
        for incr, value in enumerate(iterable, 1):
            self.parent.cell(self.row, self._max_column + incr, value)
        self.refresh(incr=len(iterable))

    def key_to_column(self, key):
        key_type = type(key)
        if key_type not in {int, str}:
            raise KeyError(f"key type should be int or str (got {type(key)}). Use str to reference by header; use int to reference by index.")
        if key_type == str:
            key = self.parent.header.get_column(key)
        return key
    
    def cell(self, key):
        '''
        Get cell by index with int key or by header with str key.
        return Cell
        '''
        key_type = type(key)
        if key_type not in {int, str}:
            raise KeyError(f"key type should be int or str (got {type(key)}). Use str to reference by header; use int to reference by index.")
        if key_type == int:
            return self.cells[key]
        else:
            column = self.parent.header.get_column(key)
            if column is None:
                raise KeyError(f"key {key} is not in header row")
            try:
                return self.cell_dict[column]
            except KeyError:
                self.refresh(self.parent.header.max_column - self.max_column)
                return self.cell_dict[column]

class Header(ContentRow):
    def __init__(self, parent, cells):
        super().__init__(parent, cells)
        self.column_map: Dict[str, int] = self._init_column_map()
        self.setup_use_default(self.use_default)
        self.default_map: Dict[int, Dict[str, Union[str, int, float, bool, None]]] = {}
    
    def rebuild(self):
        '''
        Recheck min_row and max_row of Worksheet then reset object with new cell range.
        '''
        self.cells = self.parent[self.parent.header_row]
        self._min_column = self.cells[0].column
        self._max_column = self.cells[-1].column
        self.cell_dict = self._gen_cell_dict()
        self.column_map = self._init_column_map()
        self.setup_use_default(self.use_default)

    def refresh(self, increment=0, new_row=None):
        '''
        Refresh cells; cell range remains unchanged.
        '''
        super().refresh(increment, new_row)
        self.column_map = self._init_column_map()
        self.setup_use_default(self.use_default)
    
    def __setitem__(self, key, value):
        super().__setitem__(key, value)
        self.refresh()
    
    @property
    def use_default(self):
        return self.parent.use_default
    
    def setup_use_default(self, value):
        if value != True:
            return
        row = self.row - 1
        if row <= 0:
            return
        for c in self.cells:
            cell_value = self.parent.cell(row, c.column).data
            try:
                default = eval(cell_value)
                dtype = type(default)
            except:
                default = None
                dtype = None
            self.default_map[c.column] = {
                'default': default,
                'type': dtype
            }
    
    def _init_column_map(self):
        cmap = {}
        for c in self.cells:
            cell_value = c.data
            key = str(cell_value) if cell_value is not None else ''
            if key not in cmap:
                cmap[key] = c.column
        return cmap
    
    def get_column(self, key):
        '''
        Get column by header name
        '''
        try:
            return self.column_map[key]
        except:
            return None
    
    def get_type(self, key):
        if not self.use_default:
            return None
        column = self.key_to_column(key)
        try:
            return self.default_map[column]['type']
        except:
            return None
    
    def get_default(self, key):
        if not self.use_default:
            return None
        column = self.key_to_column(key)
        try:
            return self.default_map[column]['default']
        except:
            return None


from typing import Union, List
from pycel.excelcompiler import ExcelCompiler
from openpyxl import Workbook as _Workbook
from openpyxl.chartsheet import Chartsheet
from .Worksheet import Worksheet

class Workbook(_Workbook):
    filename: str = None
    def __init__(self, filename=None, read_only=False, use_default=False):
        super().__init__(write_only=False, iso_dates=False)
        self.filename = filename
        self._read_only = read_only
        self._use_default = use_default
        self._formula_calculator = None
    
    def __getitem__(self, key) -> Union[Worksheet, Chartsheet]:
        return super().__getitem__(key)
    
    def create_sheet(self, title=None, index=None) -> Worksheet:
        new_ws = Worksheet(parent=self, title=title)
        self._add_sheet(sheet=new_ws, index=index)
        return new_ws
    
    def __enter__(self):
        return self
    
    def __exit__(self, exc_type, exc_value, traceback):
        if self._read_only or self.filename is None:
            self.close()
        else:
            self.save(self.filename)
    
    def save(self, filename=None):
        if filename is not None:
            super().save(filename)
        elif self.filename is not None:
            super().save(self.filename)
    
    def _compute(self, sheet: str, address: str):
        if self._formula_calculator is None:
            self._init_calculator()
        return self._formula_calculator.evaluate(f"{sheet}!{address}")
    
    def _init_calculator(self):
        self._formula_calculator = ExcelCompiler(excel=self)


from .open_xlsx import open_xlsx
from .new_xlsx import new_xlsx
from .Workbook import Workbook
from .Worksheet import Worksheet
from .Cell import Cell
from .ReadOnlyCell import ReadOnlyCell
from .Series import Header, ContentRow


from typing import Generator, Union, Optional
from inspect import isgenerator
from openpyxl.worksheet.worksheet import Worksheet as _Worksheet
from openpyxl.utils import (
    coordinate_to_tuple,
    column_index_from_string
)
from openpyxl.worksheet.views import (
    Pane,
    Selection,
    SheetViewList,
)
from .Series import ContentRow, Header, InvalidOperationError
from .Cell import Cell

class Worksheet(_Worksheet):
    def __init__(self, parent, title=None):
        super().__init__(parent, title)
        self._header_row: Optional[int] = None
        self._use_default: Optional[bool] = False
        self.data_only_copy: Worksheet = None
        self.header: Header = None
    
    @property
    def header_row(self):
        '''
        Which row is the header row. When header_row is not None, header is initialized.
        '''
        return self._header_row
    
    @header_row.setter
    def header_row(self, value: int):
        '''
        set the row of header and init worksheet's header
        '''
        if not isinstance(value ,int) and value is not None:
            raise TypeError(f"header_row should be int or None. (got {value}: {type(value)})")
        if isinstance(value, int) and value <= 0:
            raise ValueError(f"header_row can only be >= 1. (got {value})")
        self._header_row = value
        if value is None:
            self.header = None
        else:
            self._init_header(value)
    
    def _init_header(self, header_row):
        self.header = Header(self, self[header_row])

    @property
    def use_default(self):
        '''
        Whether use default value and type
        '''
        return self._use_default
    
    @use_default.setter
    def use_default(self, value):
        if value == True and self.header is None:
            raise ValueError(f"Cannot enable use_default cause {self.name}'s header row is not set.")
        self._use_default = value
        self.header.setup_use_default(value)
    
    def insert_rows(self, idx, amount=1):
        '''
        Insert amount of rows before the idx-th row.
        Args:
            idx: the idx-th row
            amount: amount of rows to be inserted
        '''
        if type(idx) != int:
            raise TypeError(f"idx should be of type int (got {type(idx)}.")
        if type(amount) != int:
            raise TypeError(f"idx should be of type int (got {type(amount)}.")
        if idx <= 0:
            raise ValueError(f"Cannot insert before the {idx} row.")
        if amount < 0:
            raise ValueError(f"Cannot insert {amount} rows.")
        if amount == 0:
            return
        super().insert_rows(idx, amount)
        if self.header is not None and idx <= self.header_row:
            self.header_row += amount
            self.header.refresh(new_row=self.header_row)
    
    def delete_rows(self, idx, amount=1):
        '''
        Delete amount of rows from the idx-th row (inclusive).
        Args:
            idx: the idx-th row
            amount: amount of rows to be deleted
        '''
        if type(idx) != int:
            raise TypeError(f"idx should be of type int (got {type(idx)}.")
        if type(amount) != int:
            raise TypeError(f"idx should be of type int (got {type(amount)}.")
        if idx <= 0:
            raise ValueError(f"Cannot insert before the {idx} row.")
        if amount < 0:
            raise ValueError(f"Cannot insert {amount} rows.")
        if amount == 0:
            return
        super().delete_rows(idx, amount)
        if self.header is None or self.header_row < idx:
            return
        elif self.header_row >= idx and self.header_row < idx + amount:
            self.header_row = None
            self._use_default = False
        elif self.header_row >= idx + amount:
            self.header_row -= amount
            self.header.refresh(new_row=self.header_row)
            if self.header_row == idx + amount:
                self.use_default = False
    
    def insert_cols(self, idx, amount=1):
        '''
        Insert amount of columns before the idx-th column.
        Args:
            idx: the idx-th column
            amount: amount of columns to be inserted
        '''
        if type(idx) != int:
            raise TypeError(f"idx should be of type int (got {type(idx)}.")
        if type(amount) != int:
            raise TypeError(f"idx should be of type int (got {type(amount)}.")
        if idx <= 0:
            raise ValueError(f"Cannot insert before the {idx} row.")
        if amount < 0:
            raise ValueError(f"Cannot insert {amount} rows.")
        if amount == 0:
            return
        super().insert_cols(idx, amount)
        self.header.rebuild()

    def delete_cols(self, idx, amount=1):
        '''
        Delete amount of columns from the idx-th column (inclusive).
        Args:
            idx: the idx-th column
            amount: amount of columns to be deleted
        '''
        if type(idx) != int:
            raise TypeError(f"idx should be of type int (got {type(idx)}.")
        if type(amount) != int:
            raise TypeError(f"idx should be of type int (got {type(amount)}.")
        if idx <= 0:
            raise ValueError(f"Cannot insert before the {idx} row.")
        if amount < 0:
            raise ValueError(f"Cannot insert {amount} rows.")
        if amount == 0:
            return
        super().delete_cols(idx, amount)
        self.header.rebuild()
    
    # override to keep data_only_copy in sync
    def move_range(self, cell_range, rows=0, cols=0, translate=False):
        super().move_range(cell_range=cell_range, rows=rows, cols=cols, translate=translate)
        if self.data_only_copy:
            self.data_only_copy.move_range(cell_range=cell_range, rows=rows, cols=cols, translate=translate)
        
    # override to keep data_only_copy in sync
    def _move_cell(self, row, column, row_offset, col_offset, translate=False):
        return super()._move_cell(row, column, row_offset, col_offset, translate=translate)
        if self.data_only_copy:
            self.data_only_copy._move_cell(row, column, row_offset, col_offset, translate=translate)

    def append_header(self, iterable):
        '''
        Append iterable as header row to worksheet (append a row and set sheet's row_offset).
        '''
        self.append(iterable)
        self.header_row = self._current_row
    
    def append_by_header(self, data_dict, update_header=True):
        '''
        Append to worksheet by header names. Header is auto-generated from dict keys the first time append_by_header is called.
        Args:
            data_dict: a dict containing the row data
            append_header: if True, append new header names from data_dict; else the key-value pair is ignored.
        '''
        if not isinstance(data_dict, dict):
            raise TypeError(f"append_by_header accepts only dict var (got {type(data_dict)}.")
        if self.header is None:
            names = [k for k in data_dict]
            self.append_header(names)
        converted = {}
        for k, v in data_dict.items():
            column = self.header.get_column(k)
            if column is not None:
                converted[column] = v
            elif update_header:
                self.header.append(k)
                converted[self.header.max_column] = v
        self.append(converted)
    
    @property
    def content_rows(self) -> Generator[ContentRow, None, None]:
        '''
        Return a generator of ContentRow.
        '''
        if self.header is None:
            raise InvalidOperationError(f"Cannot init content_rows when header is None. Set header_row to init header.")
        for row in self.iter_rows(self.header_row + 1):
            yield ContentRow(self, row)
    
    def get_content_column(self, key):
        '''
        Get a ContentColumn by column if key is of type int, by header if key is of type str.
        '''
        if self.header is None:
            raise InvalidOperationError("Cannot get content column when header is None. Set header_row to init header.")
        return self.header.cell(key).bottom.vertical

    def _get_cell(self, row, column):
        """
        Internal method for getting a cell from a worksheet.
        Will create a new cell if one doesn't already exist.
        """
        coordinate = (row, column)
        if not coordinate in self._cells:
            cell = Cell(self, row=row, column=column)
            self._add_cell(cell)
            return cell
        return self._cells[coordinate]

    @property
    def freeze_panes(self):
        if self.sheet_view.pane is not None:
            return self.sheet_view.pane.topLeftCell

    @freeze_panes.setter
    def freeze_panes(self, topLeftCell=None):
        if isinstance(topLeftCell, Cell):
            topLeftCell = topLeftCell.coordinate
        if topLeftCell == 'A1':
            topLeftCell = None

        if not topLeftCell:
            self.sheet_view.pane = None
            return

        row, column = coordinate_to_tuple(topLeftCell)

        view = self.sheet_view
        view.pane = Pane(topLeftCell=topLeftCell,
                        activePane="topRight",
                        state="frozen")
        view.selection[0].pane = "topRight"

        if column > 1:
            view.pane.xSplit = column - 1
        if row > 1:
            view.pane.ySplit = row - 1
            view.pane.activePane = 'bottomLeft'
            view.selection[0].pane = "bottomLeft"
            if column > 1:
                view.selection[0].pane = "bottomRight"
                view.pane.activePane = 'bottomRight'

        if row > 1 and column > 1:
            sel = list(view.selection)
            sel.insert(0, Selection(pane="topRight", activeCell=None, sqref=None))
            sel.insert(1, Selection(pane="bottomLeft", activeCell=None, sqref=None))
            view.selection = sel

    def append(self, iterable):
        """Appends a group of values at the bottom of the current sheet.

        * If it's a list: all values are added in order, starting from the first column
        * If it's a dict: values are assigned to the columns indicated by the keys (numbers or letters)

        :param iterable: list, range or generator, or dict containing values to append
        :type iterable: list|tuple|range|generator or dict

        Usage:

        * append(['This is A1', 'This is B1', 'This is C1'])
        * **or** append({'A' : 'This is A1', 'C' : 'This is C1'})
        * **or** append({1 : 'This is A1', 3 : 'This is C1'})

        :raise: TypeError when iterable is neither a list/tuple nor a dict

        """
        row_idx = self._current_row + 1

        if (isinstance(iterable, (list, tuple, range))
            or isgenerator(iterable)):
            for col_idx, content in enumerate(iterable, 1):
                if isinstance(content, Cell):
                    # compatible with write-only mode
                    cell = content
                    if cell.parent and cell.parent != self:
                        raise ValueError("Cells cannot be copied from other worksheets")
                    cell.parent = self
                    cell.column = col_idx
                    cell.row = row_idx
                else:
                    cell = Cell(self, row=row_idx, column=col_idx, value=content)
                self._cells[(row_idx, col_idx)] = cell

        elif isinstance(iterable, dict):
            for col_idx, content in iterable.items():
                if isinstance(col_idx, str):
                    col_idx = column_index_from_string(col_idx)
                cell = Cell(self, row=row_idx, column=col_idx, value=content)
                self._cells[(row_idx, col_idx)] = cell

        else:
            self._invalid_row(iterable)

        self._current_row = row_idx




# from openpyxl
import traceback
import logging
from warnings import warn

# compatibility imports
from openpyxl.xml.functions import iterparse

# package imports
from .Cell import Cell  # use reinherited Cell
from openpyxl.cell import MergedCell  # use reinherited MergedCell

from openpyxl.worksheet._reader import WorksheetReader as _WorksheetReader
from .ReadOnlyCell import ReadOnlyCell
from .WorksheetParser import WorksheetParser

logger = logging.getLogger(__name__)

class WorksheetReader(_WorksheetReader):
    """
    Create a parser and apply it to a workbook
    """
    def __init__(self, ws, xml_source, shared_strings, data_only, read_only):
        self.ws = ws
        self.parser = WorksheetParser(xml_source, shared_strings, data_only, ws.parent.epoch, ws.parent._date_formats)
        self.tables = []
        self.read_only = read_only

    def bind_cells(self):
        for idx, row in self.parser.parse():
            for cell in row:
                # logger.debug(cell)
                style = self.ws.parent._cell_styles[cell['style_id']]
                c = Cell(self.ws, row=cell['row'], column=cell['column'], style_array=style)
                c._value = cell['value']
                c.data_type = cell['data_type']
                # read cache of formula cell
                if c.data_type == 'f':
                    c._cache = cell['cache']
                    c._cache_type = cell['cache_type']
                self.ws._cells[(cell['row'], cell['column'])] = c
        self.ws.formula_attributes = self.parser.array_formulae
        if self.ws._cells:
            self.ws._current_row = self.ws.max_row # use cells not row dimensions
    
    def bind_cells_read_only(self):
        for idx, row in self.parser.parse():
            for cell in row:
                c = ReadOnlyCell(self.ws, **cell)
                self.ws._cells[(cell['row'], cell['column'])] = c
        self.ws.formula_attributes = self.parser.array_formulae
        if self.ws._cells:
            self.ws._current_row = self.ws.max_row # use cells not row dimensions

    def bind_all(self):
        if self.read_only:
            self.bind_cells_read_only()
        else:
            self.bind_cells()
        self.bind_merged_cells()
        self.bind_hyperlinks()
        self.bind_formatting()
        self.bind_col_dimensions()
        self.bind_row_dimensions()
        self.bind_tables()
        self.bind_properties()



from setuptools import setup, find_packages
from importlib.util import  module_from_spec, spec_from_file_location

spec = spec_from_file_location("constants", "./pyxlsx/_constants.py")
constants = module_from_spec(spec)
spec.loader.exec_module(constants)

with open('README.md', 'r') as fp:
    long_description = fp.read()

__author__ = constants.__author__
__url__ = constants.__url__
__version__ = constants.__version__
__license__ = constants.__license__

setup(
    name='pyxlsx',
    packages=find_packages(
        exclude=['*.tests', '*.test_.*', 'tests', 'develop']
    ),
    package_dir={},
    # metadata
    author=__author__,
    long_description=long_description,
    long_description_content_type='text/markdown',
    url=__url__,
    version=__version__,
    license=__license__,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.5',
    install_requires=[
        "openpyxl>=3.0.3",
        "pycel"
    ]
)


import os
import logging
from .Workbook import Workbook
from .ExcelReader import ExcelReader

logger = logging.getLogger(__name__)

def open_xlsx(filename, read_only=False) -> Workbook:
    '''
    Open an existing xlsx file.
    Args: 
        filename: path of a xlsx file
        read_only: whether the xlsx file can only be read
    Returns:
        Workbook
    '''
    if not os.path.exists(filename):
        raise FileNotFoundError(f"{filename} not found.")
    reader = ExcelReader(filename, read_only=read_only)
    reader.read()
    wb = reader.wb
    wb.filename = filename
    # data_only copy is opened for reading cached formula results
    # reader_data_only = ExcelReader(filename, read_only=read_only, data_only=True)
    # reader_data_only.read()
    # wb_data_only = reader_data_only.wb
    # for name in wb.sheetnames:
    #     wb[name].data_only_copy = wb_data_only[name]
    return wb


from openpyxl.reader.workbook import WorkbookParser as _WorkbookParser
from .Workbook import Workbook

class WorkbookParser(_WorkbookParser):
    def __init__(self, archive, workbook_part_name, keep_links=True):
        self.archive = archive
        self.workbook_part_name = workbook_part_name
        self.wb = Workbook()
        self.keep_links = keep_links
        self.sheets = []


from .Workbook import Workbook

def new_xlsx(filename=None) -> Workbook:
    '''
    Create a new xlsx file.
    Args:
        filename: path to save the new xlsx file.
    Returns:
        Workbook
    '''
    wb = Workbook(filename)
    return wb


import pytest
from pathlib import Path
from openpyxl import load_workbook

@pytest.fixture
def data_path(request):
    return Path(r"/Users/fort/git/python_projects/pyxlsx/pyxlsx/tests/data")


def test_read(data_path):
    path = data_path / 'cache_test_1.xlsx'
    wb = load_workbook(path, data_only=False)
    ws = wb.active
    for row in ws.rows:
        for c in row:
            print(c.value, end=' ')
        print()



import logging
from warnings import warn
from openpyxl.utils import (
    get_column_letter,
    coordinate_to_tuple
)
from openpyxl.cell.text import Text
from openpyxl.utils.datetime import (
    from_excel, from_ISO8601, WINDOWS_EPOCH
)
from openpyxl.worksheet._reader import (
    CELL_TAG, VALUE_TAG, FORMAT_TAG, MERGE_TAG, INLINE_STRING,
    COL_TAG, ROW_TAG, CF_TAG, LEGACY_TAG, PROT_TAG, EXT_TAG, 
    HYPERLINK_TAG, TABLE_TAG, PRINT_TAG, HEADER_TAG, FILTER_TAG,
    VALIDATION_TAG, PROPERTIES_TAG, VIEWS_TAG, FORMULA_TAG, 
    ROW_BREAK_TAG, COL_BREAK_TAG, SCENARIOS_TAG, DATA_TAG, 
    DIMENSION_TAG, CUSTOM_VIEWS_TAG,
    _cast_number
)
from openpyxl.worksheet._reader import WorkSheetParser as _WorksheetParser

logger = logging.getLogger(__name__)

class WorksheetParser(_WorksheetParser):
    def parse_cell(self, element):
        data_type = element.get('t', 'n')
        coordinate = element.get('r')
        self.col_counter += 1
        style_id = element.get('s', 0)
        if style_id:
            style_id = int(style_id)

        if data_type == "inlineStr":
            value = None
        else:
            value = element.findtext(VALUE_TAG, None) or None

        if coordinate:
            row, column = coordinate_to_tuple(coordinate)
        else:
            row, column = self.row_counter, self.col_counter

        # logger.debug(f"after parse: ({row}, {column}) {data_type}: {value}")

        if element.find(FORMULA_TAG) is not None:
            cache_type = data_type
            cache = value
            # logger.debug(f"cache before parse {cache_type}: {cache}")
            data_type = 'f'
            value = self.parse_formula(element)
            cache_type, cache = self._parse_value(element, cache_type, cache, style_id)
            # logger.debug(f"cache before parse {cache_type}: {cache}")
            # logger.debug(f"formula after parse: ({row}, {column}) {data_type}: {value}")
            return {
                'row':row, 
                'column':column, 
                'value':value, 
                'data_type':data_type, 
                'style_id':style_id,
                'cache_type': cache_type,
                'cache': cache
            }

        else:
            data_type, value = self._parse_value(element, data_type, value, style_id)
            # logger.debug(f"after parse: ({row}, {column}) {data_type}: {value}")
            return {'row':row, 'column':column, 'value':value, 'data_type':data_type, 'style_id':style_id}

    def _parse_value(self, element, data_type, value, style_id):
        if value is not None:
            if data_type == 'n':
                value = _cast_number(value)
                if style_id in self.date_formats:
                    data_type = 'd'
                    try:
                        value = from_excel(value, self.epoch)
                    except ValueError:
                        msg = """Cell {0} is marked as a date but the serial value {1} is outside the limits for dates. The cell will be treated as an error.""".format(coordinate, value)
                        warn(msg)
                        data_type = "e"
                        value = "#VALUE!"
            elif data_type == 's':
                value = self.shared_strings[int(value)]
            elif data_type == 'b':
                value = bool(int(value))
            elif data_type == "str":
                data_type = "s"
            elif data_type == 'd':
                value = from_ISO8601(value)

        elif data_type == 'inlineStr':
                child = element.find(INLINE_STRING)
                if child is not None:
                    data_type = 's'
                    richtext = Text.from_tree(child)
                    value = richtext.content

        return (data_type, value)
        